from ipywidgets import widgets,interact, Layout
import rwth_nb.plots.colors as rwthcolors
rwth_colors = rwthcolors.rwth_colors

import datetime
import json
import os
import hashlib # for anonymizing the username

def get_notebook_name():
    try:
        # import necessary packages (see https://github.com/jupyter/notebook/issues/1000)
        # importing inside function because those libraries should only be included if they are realy needed
        import json
        import re
        import ipykernel
        import requests
        import os

        from requests.compat import urljoin
        from notebook.notebookapp import list_running_servers

        # now determine calling notebook name
        kernel_id = re.search('kernel-(.*).json',
                              ipykernel.connect.get_connection_file()).group(1)
        servers = list_running_servers()
        for ss in servers:
            response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                    params={'token': ss.get('token', '')})
            for nn in json.loads(response.text):
                if nn['kernel']['id'] == kernel_id:
                    relative_path = nn['notebook']['path']
                    return os.path.splitext(os.path.split(os.path.join(ss['notebook_dir'], relative_path))[1])[0]
    except:
        return 'tmp'


def rwth_feedback(feedback_name, questions, feedback_path='feedback.json', mail_to=None, mail_from='feedback@jupyter.rwth-aachen.de', mail_subject=None, mail_smtp_host='smarthost.rwth-aachen.de'):
    global widgets_container
    is_feedback_edited = is_send_confirmed = False

    def on_feedback_edited(_):
        nonlocal is_feedback_edited
        is_feedback_edited = True

    # Internationalization
    feedback_scale_options_de = ['Stimme voll zu', 'Ich stimme zu', 'Keine Meinung', 'Ich stimme nicht zu', 'Ich stimme gar nicht zu'] # Likert-scale
    feedback_text_de = {"your-feedback" : "Dein Feedback ...",
                    "send" : "Absenden",
                    "confirm_send" : "Zum Absenden bitte bestätigen.",
                    "sent" : "Dein Feedback wurde abgeschickt. Vielen Dank!",
                    "empty" : "Bitte geben Sie zuerst Feedback ein, das abgesendet werden kann.",
                    "mailfailure": "Die Mail mit dem Feedback konnte nicht versendet werden. Das Feedback wurde lokal abgespeichert."}
    # TODO: Add at least English here
    
    # Select language
    feedback_scale_options = feedback_scale_options_de; feedback_text = feedback_text_de;
    
    # Default arguments for toggle_button and textarea
    toggle_args = {"options" : feedback_scale_options, 
                   "index" : 2,  "description" : "",  "disabled" : False,  "style": {'button_color': rwth_colors['rwth:green-50']},  "tooltips" : []}
    textarea_args = {"value" : "",  "placeholder" : feedback_text['your-feedback'], 
                     "description" : "",  "disabled" : False}
    
    widgets_container = [];
    widgets_values = [];
    for question in questions:
        if question['type'] == 'free-text':
             # Free text
            widget_label = widgets.HTML(value="<b>{}</b>".format(question['label']))
            widget_value = widgets.Textarea(**textarea_args)
            widget_value.observe(on_feedback_edited, 'value')
                        
        elif question['type'] == 'scale':
            # Toggle Buttons
            widget_label = widgets.HTML(value="<b>{}</b>".format(question['label']))
            widget_value = widgets.ToggleButtons(**toggle_args)
            widget_value.observe(on_feedback_edited, 'value')

        widgets_container.append(widget_label)
        widgets_container.append(widget_value)
        widgets_values.append(widget_value) # TODO: Get rid of this?

    # Button 
    button = widgets.Button(description = feedback_text['send'], disabled = False, style= {'button_color': rwth_colors['rwth:green-50'], 'margin':'10px'}, icon='', layout=Layout(margin='20px 0 0 0'))
    output = widgets.Output()
    
    # Button click callback
    def on_button_clicked(b):
        nonlocal is_send_confirmed

        if not button.disabled:
            entry = {}

            nonlocal is_feedback_edited
            if not is_feedback_edited:
                with output:
                    print(feedback_text['empty'])    
            else:
                # set up json entries
                entry['name'] = feedback_name
                entry['date'] = "{}".format(datetime.datetime.now())
                if 'USER' in os.environ.keys():
                    user_name = os.environ['USER']
                else:
                    user_name = os.environ['HOSTNAME']  # take hostname as fall-back
                entry['userhash'] = hashlib.sha256(str.encode(user_name)).hexdigest()
                entry['answer'] = {q['id']: v.value for q, v in zip(questions, widgets_values)}

                if not is_send_confirmed:
                    button.description = feedback_text['confirm_send']
                    button.layout.width = 'auto'
                    is_send_confirmed = True
                    return

                # dump entries into json file
                if not os.path.isfile(feedback_path):
                    with open(feedback_path, mode='w', encoding='utf-8') as f:
                        json.dump([], f)
                with open(feedback_path, mode='r', encoding='utf-8') as f:
                    entries = json.load(f)
                with open(feedback_path, mode='w', encoding='utf-8') as f:
                    entries.append(entry)
                    json.dump(entries, f)

                # send json file as attachment of mail
                if mail_to is not None:
                    try:
                        import smtplib
                        from email.mime.multipart import MIMEMultipart
                        from email.mime.base import MIMEBase
                        from email import encoders
                        
                        # create message                
                        msg = MIMEMultipart()
                        msg['From'] = mail_from
                        msg['To'] = mail_to
                        
                        if mail_subject is not None:
                            msg['Subject'] = mail_subject
                        else:
                            msg['Subject'] = feedback_name

                        # open the file to be sent as attachment
                        with open(feedback_path, 'rb') as attachment:
                            # attach file to mail
                            p = MIMEBase('application', 'octet-stream')
                            p.set_payload(attachment.read())

                        # encode and add as attachment to message
                        encoders.encode_base64(p)
                        p.add_header('Content-Disposition', "attachment; filename= %s" % feedback_path)
                        msg.attach(p)

                        # send mail
                        s = smtplib.SMTP(mail_smtp_host)
                        text = msg.as_string()
                        s.sendmail(mail_from, mail_to, text)

                        # close connection
                        s.quit()

                        with output:
                            print(feedback_text['sent'])

                    except ConnectionRefusedError:
                        # Not connected to the RWTH network
                        with output:
                            print(feedback_text['mailfailure'])

                button.disabled = True

    # Set callback to button
    button.on_click(on_button_clicked)

    # Display widgets
    display(widgets.VBox(widgets_container), 
        button, output)


from scipy import signal  # butter


def butter(cutoff, fs, order=5, type='Tiefpass', fdelta=0):
    """Butterworth Filter of order n

        Parameters
        ----------
        cutoff : float
            cutoff frequency
        fs : float
            sampling frequency, is used to calculate nyquist frequency
        order : float, optional
            order of filter
        type: {'Tiefpass', 'Bandpass', 'Hochpass'}, optional
            type of filter.
        fdelta: float, optional
            bandwith of filter


        Returns
        -------
        b : ndarray
            numerator polynomial of the IIR filter
        a : ndarray
            denominator polynomial of the IIR filter

        """
    # bandpass type
    btypes = {'Tiefpass': 'lowpass', 'Bandpass': 'bandpass', 'Hochpass': 'highpass'}
    btype = btypes[type]

    # Nyquist frequency
    nyq = 0.5 * fs

    if btype == 'bandpass':
        # normalized min and max frequency
        normal_min = (cutoff-fdelta/2) / nyq
        normal_max = (cutoff+fdelta/2) / nyq
        # normalized cutoff-frequency
        normal_cutoff = [normal_min, normal_max]
    else:
        # normalized cutoff-frequency
        normal_cutoff = cutoff / nyq
    b, a = signal.butter(order, normal_cutoff, btype=btype, analog=False)  # coefficients in z-domain
    return b, a


# Shortcuts
def butter_bandpass(f0, fdelta, fs, order=5):
    """Bandpass Butterworth Filter

    This filter is set as a bandpass filter with f0, fdelta, fs and order set by user.

    See Also
    --------
    butter : Filter design using order and critical points

    """
    b, a = butter(f0, fs, order, 'Bandpass', fdelta)
    return b, a


def butter_lowpass(cutoff, fs, order=5):
    """Lowpass Butterworth Filter

    This filter is set as a lowpass filter with cutoff, fs and order set by user.

    See Also
    --------
    butter : Filter design using order and critical points

    """
    b, a = butter(cutoff, fs, order, 'Tiefpass')
    return b, a


def butter_highpass(cutoff, fs, order=5):
    """Highpass Butterworth Filter

    This filter is set as a highpass filter with cutoff, fs and order set by user.

    See Also
    --------
    butter : Filter design using order and critical points

    """
    b, a = butter(cutoff, fs, order, 'Hochpass')
    return b, a


def filter(s, b, a):
    """Digital Filter
        
        Filter s(n) in z-Domain with filter coefficients a and b:
                            -1              -M
                b[0] + b[1]z  + ... + b[M] z
        G(z) = -------------------------------- S(z)
                            -1              -N
                a[0] + a[1]z  + ... + a[N] z

        Parameters
        ----------
        s : array_like
            n-dimensional input array
        b : array_like
            numerator coefficient vector in a 1-D sequence.
        a : array_like
            denominator coefficient vector in a 1-D sequence.

        Returns
        -------
        g : array
            output of digital filter.
    """
    g = signal.lfilter(b, a, s)
    return g



"""
This module defines functionality related to signal processing.
"""

import numpy as np
from scipy.signal import find_peaks # peak finder

# t axis in seconds
(t, deltat) = np.linspace(-10, 10, 5001, retstep=True)  

# Elementary signals
gauss     = lambda t: np.exp(-(t)**2)
unitstep  = lambda t: np.where(t>=0, 1, 0)
rect      = lambda t: unitstep(t+0.5) - unitstep(t-0.5)
tri       = lambda t: rect(t/2)*(1-abs(t))
si        = lambda t: np.sinc(t/np.pi) # English notation sinc(t) = sin(pi t)/(pi t)
rc_tp     = lambda f: 1 / (1 + 1j * 2 * np.pi * f) # RC = 1
rl_hp     = lambda f: (1j * 2 * np.pi * f) / (1 + 1j * 2 * np.pi * f) # RL = 1

eps = np.finfo(float).eps # floating point accuracy



def dft(s, fs, NFFT=0):
    """
    Calculate discrete Fourier transform of vector s
    Sampling frequency fs is used to calculate frequency vector f
    Number of frequency coefficients can be specified as well
    """
    if NFFT==0: 
        NFFT = len(s) # int(2**np.ceil(np.log2(len(bla))))
    
    S = np.fft.fftshift(np.fft.fft(s, NFFT))/NFFT
    f = np.linspace(-fs/2,fs/2, NFFT)
    
    return S, f


def idft(S, Ntime=0, NFFT=0):
    """
    Calculate discrete inverse Fourier transform of vector S
    Number of time bins is used to crop the output of NumPy's ifft function
    Number of frequency coefficients can be specified as well
    """
    if NFFT == 0: NFFT = len(S)
    
    s = np.fft.ifft(np.fft.ifftshift(S*NFFT), NFFT)
    
    if not Ntime == 0:
        s = s[0:Ntime]
    
    return s


def find_intervals(s, t, thresh, delta):
    """
    Find intervals of signal s by searching for delta-functions in the second derivative of s

    Parameters
    ----------
    s : array_like
        The signal
        
    t : array_like
        Corresponding t-axis
        
    thresh : float
        Threshold for delta search
        
    delta : float
        Sampling period

    Returns
    -------
    intervals_s, peaks, dd : *intervals* are the intervals, *peaks* the found peaks and *dd* the second derivative of s.
    """
        
    # derivative of derivative shows delta functions at discontinuities
    dd = np.diff(np.diff(s, prepend=0), prepend=0)/delta**2 
    
    # find delta functions
    peaks, _  = find_peaks(np.abs(dd), prominence=thresh) 
    
    # return interval limits in seconds
    intervals = np.round(t[peaks]*10)/10
    
    return intervals, peaks, dd


def find_ind_least_diff(x, x0):
    """
    Find index of the value that's nearest to x0

    Parameters
    ----------
    x : array_like
        Array to be inspected

    x0 : float
        Target value to be searched for

    Returns
    -------
    index : int or list
        index(indices) of value(s) that is (are) nearest to x0
    """
    
    if isinstance(x0, list) or isinstance(x0, np.ndarray):
        return list(map(lambda t0: np.abs(x - t0).argmin(), x0))
    else:
        return np.abs(x - x0).argmin()




import numpy as np
from scipy.signal import residue, residuez
from scipy.special import factorial
from rwth_nb.misc.signals import eps, find_ind_least_diff, unitstep


def dft(s, fs, NFFT=0):
    """Calculate discrete fourier transform of vector s

        Parameters
        ----------
        s : array_like
            vector to be transformed
        fs : float
            sampling frequency, is used to calculate frequency vector f
        NFFT : float, optional
            number of frequency coefficients

        Returns
        -------
        S : ndarray
            resulting discrete fourier transform
        f : ndarray
            frequency vector

        Examples
        --------
        For examples see primer at rwth_nb/RWTH Transforms.ipynb
        """
    if NFFT==0: 
        NFFT = len(s) # int(2**np.ceil(np.log2(len(bla))))
    
    S = np.fft.fftshift(np.fft.fft(s, NFFT))/NFFT
    f = np.linspace(-fs/2,fs/2, NFFT)
    
    return S, f


def idft(S, Ntime=0, NFFT=0):
    """Calculate inverse discrete fourier transform of vector S

        Parameters
        ----------
        S : array_like
            vector to be transformed
        Ntime : float, optional
            number of time bins, is used to crop the output of NumPy's ifft function
        NFFT : float, optional
            number of frequency coefficients

        Returns
        -------
        s : ndarray
            resulting inverse discrete fourier transform

        Examples
        --------
        For examples see primer at rwth_nb/RWTH Transforms.ipynb
    """
    if NFFT == 0: NFFT = len(S)
    
    s = np.fft.ifft(np.fft.ifftshift(S*NFFT), NFFT)
    
    if not Ntime == 0:
        s = s[0:Ntime]
    
    return s


def ilaplace_ht(t=np.linspace(-6, 6, num=1024), H0=1, pp=np.array([]), pz=np.array([]), ord_p=np.array([]),
                     ord_z=np.array([]), roc=[-12, 12]):
    """Calculate inverse laplace transform h(t) of H(p) defined by its gain factor, poles and zeroes

        Parameters
        ----------
        t : array_like
            array of t-domain to be transformed to
        H0 : float
            gain factor
        pp : array_like
            poles on pz-plane
            (exclude conjugated poles)
        pz : array_like
            zeroes on pz-plane
            (exclude conjugated zeroes)
        ord_p : array_like
            poles' orders
        ord_z : array_like
            zeroes' orders
        roc : array_like
            region of convergence
            (range from -infinity to infinity)

        Returns
        -------
        h : ndarray
            calculated inverse Laplace-transform according to t-domain
        td: ndarray
            dirac's x coordinate if existing, empty otherwise
        sd: ndarray
            dirac's gain factor if existing, empty otherwise

        Examples
        --------
        Calculate inverse Laplace-transform with gain H0 = 1
            poles:
                p_p1 = -3 (order 1)
                p_p2 = +1 (order 2)
            zeroes:
                p_n1 = 0 (order 1)
            region of convergence:
                1 to infinity

        >>> from rwth_nb.misc.transforms import ilaplace_ht
        >>>
        >>> t = numpy.linspace(-5, 5, 1024)
        >>> H0 = 1
        >>> poles = [-3, 1]
        >>> poles_order = [1, 2]
        >>> zeroes = [0]
        >>> zeroes_order = [1]
        >>> roc = [1, numpy.inf]
        >>>
        >>> h_t, td, sd = ilaplace_ht(t, H0, poles, zeroes, poles_order, zeroes_order, roc)

        For more examples see primer at rwth_nb/RWTH Transforms.ipynb
        """
    # check for correct input
    pp = np.array(pp)
    pz = np.array(pz)
    ord_p = np.array(ord_p)
    ord_z = np.array(ord_z)

    if pp.size != ord_p.size or pz.size != ord_z.size:
        raise ValueError('Poles/Zeroes array must be of the same size as their orders.')

    if np.any(np.logical_and(roc[0] < np.real(pp), np.real(pp) < roc[1])):
        raise ValueError('The ROC cannot contain any poles.')

    # declare arrays for dirac delta function
    td = sd = np.array([])

    # append conjugated poles and zeroes and their orders to existing arrays
    poles = np.array([])
    for ind, p in enumerate(pp):
        for i in range(0, ord_p[ind]):
            poles = np.append(poles, p)

    zeroes = np.array([])
    for ind, p in enumerate(pz):
        for i in range(0, ord_z[ind]):
            zeroes = np.append(zeroes, p)

    poles = np.append(poles, np.conj(poles[np.where(poles.imag != 0)]))
    zeroes = np.append(zeroes, np.conj(zeroes[np.where(zeroes.imag != 0)]))

    # configure numerator and denominator for partial fraction
    numerator = H0 * np.poly(zeroes)
    denominator = np.poly(poles)

    if poles.size >= zeroes.size:

        r, p, k = residue(numerator, denominator)

        # round decimals
        r = np.around(r, 5)
        p = np.around(p, 5)

        # determine (anti-)causal poles
        causal = np.real(p) <= roc[0]

        # alpha = -p
        p = -p

        # inverse laplace transform
        h = np.zeros(t.shape)
        n = 1
        for ind, alpha in enumerate(p):
            # determine order
            if ind > 0:
                if p[ind] == p[ind - 1]:
                    n += 1
                else:
                    n = 1

            coeff = r[ind] * np.exp(-alpha * t) * t ** (n - 1) / factorial(n - 1)

            if causal[ind]:  # if pole is causal
                tmp = coeff * unitstep(t)
            else:
                tmp = (-1) * coeff * unitstep(-t - 0.0001)

            h = np.add(h, tmp)

        # add rest as k[0] * delta(t)
        if k.size > 0:
            td = np.array([0])
            sd = np.array([k[0]])

    else:
        # Zaehlergrad größer als Nennergrad. Keine Partialbruchzerlegung möglich
        h = np.ones(t.shape) * np.nan

    return h, td, sd


def ilaplace_Hf(f=np.linspace(-6, 6, num=1024), H0=1, pp=np.array([]), pz=np.array([]), ord_p=np.array([]),
                     ord_z=np.array([]), dB=False):
    """Calculate frequency response H(f) of H(p) defined by its gain factor, poles and zeroes

        Parameters
        ----------
        f : array_like
            array of f-domain to be transformed to
        H0 : float
            gain factor
        pp : array_like
            poles on pz-plane
            (exclude conjugated poles)
        pz : array_like
            zeroes on pz-plane
            (exclude conjugated zeroes)
        ord_p : array_like
            poles' orders
        ord_z : array_like
            zeroes' orders
        dB : bool
            return frq response in dB if true

        Returns
        -------
        _ : ndarray
            calculated frequency response (in dB if parameter dB is true)

        Examples
        --------
        Calculate frequency response of H(p) with gain H0 = 1
            poles:
                p_p1 = +j, -j (order 1) Note: Exclude conjugated poles!
            zeroes:
                p_n1 = 0 (order 1)
        and return in dB

        >>> from rwth_nb.misc.transforms import ilaplace_Hf
        >>>
        >>> f = numpy.linspace(-6, 6, 1024)
        >>> H0 = 1
        >>> poles = [1j]  # Exclude conjugated pole
        >>> poles_order = [1]
        >>> zeroes = [0]
        >>> zeroes_order = [1]
        >>> dB = True
        >>>
        >>> H_f = ilaplace_Hf(f, H0, poles, zeroes, poles_order, zeroes_order, dB)

        For more examples see primer at rwth_nb/RWTH Transforms.ipynb
        """

    numerator = H0 * np.ones(f.shape)

    for ind, ppz in enumerate(pz):
        for _ in range(1, ord_z[ind] + 1):
            numerator = numerator * (1j * f - ppz)
            if np.abs(np.imag(ppz)):
                numerator = numerator * (1j * f - np.conj(ppz))

    denominator = np.ones(f.shape)

    for ind, ppp in enumerate(pp):
        for _ in range(1, ord_p[ind] + 1):
            denominator = denominator * (1j * f - ppp)
            if np.abs(np.imag(ppp)):
                denominator = denominator * (1j * f - np.conj(ppp))
    if dB:
        return 20 * np.log10(np.maximum(eps, np.abs(numerator / denominator)))
    else:
        return np.abs(numerator / denominator)


def iz_hn(n=np.linspace(-6, 6, num=13), H0=1, pp=np.array([]), pz=np.array([]), ord_p=np.array([]),
               ord_z=np.array([]), roc=[0, 12]):
    """Calculate inverse z-transform h(n) of H(z) defined by its gain factor, poles and zeroes

    Parameters
    ----------
    n : array_like
        array of n-domain to be transformed to
    H0 : float
        gain factor
    pp : array_like
        poles on pz-plane
        (exclude conjugated poles)
    pz : array_like
        zeroes on pz-plane
        (exclude conjugated zeroes)
    ord_p : array_like
        poles' orders
    ord_z : array_like
        zeroes' orders
    roc : array_like
        region of convergence
        (range from 0 to infinity)

    Returns
    -------
    h : ndarray
        calculated inverse z-transform according to n-domain

    Examples
    --------
    Calculate inverse z-transform with gain H0 = 1
        poles:
            z_p1 = +0.5 (order 1)
            z_p2 = +1 (order 2)
        zeroes:
            z_n1 = +2 (order 1)
        region of convergence:
            0.5 to 1

    >>> from rwth_nb.misc.transforms import iz_hn
    >>>
    >>> n = [-3, -2, -1, 0, 1, 2, 3]
    >>> H0 = 1
    >>> poles = [0.5, 1]
    >>> poles_order = [1, 2]
    >>> zeroes = [2]
    >>> zeroes_order = [1]
    >>> roc = [0.5, 1]
    >>>
    >>> h_n = iz_hn(n, H0, poles, zeroes, poles_order, zeroes_order, roc)

    """
    # check for correct input
    pp = np.array(pp)
    pz = np.array(pz)
    ord_p = np.array(ord_p)
    ord_z = np.array(ord_z)

    if np.isinf(roc[0]):
        raise ValueError('Region of convergence must start at 0 or pole (=[0, b] anti-causal or =[b, inf] causal).')

    if pp.size != ord_p.size or pz.size != ord_z.size:
        raise ValueError('Poles/Zeroes array must be of the same size as their orders.')

    # append conjugated poles and zeroes and their orders to existing arrays
    poles = np.array([])
    for ind, p in enumerate(pp):
        for i in range(0, ord_p[ind]):
            poles = np.append(poles, p)

    zeroes = np.array([])
    for ind, p in enumerate(pz):
        for i in range(0, ord_z[ind]):
            zeroes = np.append(zeroes, p)

    poles = np.append(poles, np.conj(poles[np.where(poles.imag != 0)]))
    zeroes = np.append(zeroes, np.conj(zeroes[np.where(zeroes.imag != 0)]))

    if poles.size >= zeroes.size:
        # configure numerator and denominator for partial fraction
        numerator = H0 * np.poly(zeroes)
        denominator = np.poly(poles)
        for i in range(0, int(poles.size - zeroes.size)):
            numerator = np.insert(numerator, 0, 0)

        # calculate residue
        r, p, k = residuez(numerator, denominator)

        # round decimals
        r = np.around(r, 5)
        p = np.around(p, 5)

        # determine (anti-)causal poles
        causal = np.around(np.abs(p), 5) <= np.abs(roc[0])

        # inverse z-transform
        h = np.zeros(n.shape)
        m = 1
        for ind, b in enumerate(p):
            # determine order
            if ind > 0:
                if p[ind] == p[ind-1]:
                    m += 1
                else:
                    m = 1

            coeff = 1
            for i in range(1, m):
                coeff *= (n + i) / factorial(m - 1)
            coeff = r[ind] * coeff * b**n

            if causal[ind]:  # if pole is causal
                tmp = coeff * unitstep(n)
            else:
                tmp = (-1) * coeff * unitstep(-n - m)

            h = np.add(h, tmp)

        # add k[i] * z**(-i) as k[i] * delta(n-i) to h(n)
        if k.size > 0:
            for ind, kk in enumerate(k):
                h = np.add(h, kk * np.where(n == ind, 1, 0))
    else:
        # partial fraction not possible: return nan array
        h = np.ones(n.shape) * np.nan

    return h


def iz_Hf(f=np.linspace(-6, 6, num=1024), H0=1, pp=np.array([]), pz=np.array([]), ord_p=np.array([]),
                     ord_z=np.array([]), dB=False):
    """Calculate frequency response H(f) of z-transformed H(z) defined by its gain factor, poles and zeroes

        Parameters
        ----------
        f : array_like
            array of f-domain to be transformed to
        H0 : float
            gain factor
        pp : array_like
            poles on pz-plane
            (exclude conjugated poles)
        pz : array_like
            zeroes on pz-plane
            (exclude conjugated zeroes)
        ord_p : array_like
            poles' orders
        ord_z : array_like
            zeroes' orders
        dB : bool
            return frq response in dB if true

        Returns
        -------
        _ : ndarray
            calculated frequency response (in dB if parameter dB is true)

        Examples
        --------
        Calculate frequency response of H(z) with gain H0 = 1
            poles:
                z_p1/2 = +j, -j (order 1) Note: Exclude conjugated poles!
            zeroes:
                z_n1 = 0 (order 1)
        and return in dB

        >>> from rwth_nb.misc.transforms import iz_Hf
        >>>
        >>> f = numpy.linspace(-6, 6, 1024)
        >>> H0 = 1
        >>> poles = [1j]  # Exclude conjugated pole
        >>> poles_order = [1]
        >>> zeroes = [0]
        >>> zeroes_order = [1]
        >>> dB = True
        >>>
        >>> H_f = iz_Hf(f, H0, poles, zeroes, poles_order, zeroes_order, dB)

        For more examples see primer at rwth_nb/RWTH Transforms.ipynb
        """
    numerator = H0 * np.ones(f.shape)

    for ind, ppz in enumerate(pz):
        for _ in range(1, ord_z[ind] + 1):
            numerator = numerator * (np.exp(1j * 2 * np.pi * f) - ppz)
            if np.abs(np.imag(ppz)):
                numerator = numerator * (np.exp(1j * 2 * np.pi * f) - np.conj(ppz))

    denominator = np.ones(f.shape)

    for ind, ppp in enumerate(pp):
        for _ in range(1, ord_p[ind] + 1):
            denominator = denominator * (np.exp(1j * 2 * np.pi * f) - ppp)
            if np.abs(np.imag(ppp)):
                denominator = denominator * (np.exp(1j * 2 * np.pi * f) - np.conj(ppp))
    if dB:
        return 20 * np.log10(np.maximum(eps, np.abs(numerator / denominator)))
    else:
        return np.abs(numerator / denominator)


def ideal_sample(x, y, T):
    def x_mirrored_around_zero(maxval, inc=1):
        x = np.arange(inc, maxval + inc, inc)
        return np.r_[-x[::-1], 0, x]
    xf = x_mirrored_around_zero(np.amax(x), T)
    yf = y[find_ind_least_diff(x, xf)]
    return xf, yf


def real_sample(x, y, T, T0):
    yf = np.zeros(x.shape)
    indices = np.where(np.mod(x + T0 / 2, T) <= T0)
    yf[indices[0]] = y[indices[0]]
    return x, yf


def sample(x, y, T, T0=0):
    if T0:
        return real_sample(x, y, T, T0)
    else:
        return ideal_sample(x, y, T)



"""
Plot functions enhancing Matplotlib
"""
from matplotlib import rcParams

rcParams["axes.axisbelow"] = False

# rcParams['font.family'] = 'sans-serif'
# rcParams['font.sans-serif'] = ['Arial'] # TODO
# rcParams['font.size'] = 14
# rcParams['text.usetex'] = True
# rcParams['text.latex.unicode'] = True

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.gridspec as gridspec
import matplotlib.transforms as trans
import matplotlib.colors as mcolors

from rwth_nb.plots import colors
import numpy as np


# Propagate rwth_colors to default matplotlib colors
mcolors.get_named_colors_mapping().update(colors.rwth_colors)

# set rwth color cycle
from cycler import cycler
rcParams["axes.prop_cycle"] = cycler(color=['rwth:blue', 'rwth:orange', 'rwth:green', 'rwth:red', 'rwth:purple',
                                            'rwth:bordeaux', 'rwth:violet',  'rwth:black-50', 'rwth:maigrun',
                                            'rwth:turquoise'])


def axis(ax):
    """
    Decorate axes RWTH way.
    
    Parameters
    ----------
    ax: matplotlib.axes.Axes
        Axes to be beautified


    See Also
    --------
    TODO
    
    Notes
    -----
    Nada.

    Examples
    --------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> import rwth_nb.plot.mpl_decorations as rwth_plt
    >>>
    >>> x = np.linspace(-5,5,100)
    >>> fig,ax = plt.subplots()
    >>> ax.plot(x, x**2)
    >>> ax.set_xlabel('x'); ax.set_ylabel('f(x)');
    >>> rwth_plt.axis(ax)
    """
    ax.spines['right'].set_color('none')
    ax.spines['top'].set_color('none')

    def on_xlims_change(ax):
        # update x spine position to be always at zero, left or right according to set limits
        # update y-label x-position and horizontal alignment
        if (np.array(ax.get_xlim()) < 0).all():  # all limits negative
            left_spine_pos = ('axes', 1)
            ylabel_xpos = 1
            ylabel_halignment = 'right'
        elif (np.array(ax.get_xlim()) > 0).all():  # all limits positive
            left_spine_pos = ('axes', 0)  # spine at the left
            ylabel_xpos = 0
            ylabel_halignment = 'left'
        else:  # zero is in plot
            left_spine_pos = 'zero'  # spine at zero ([0, 0])
            xmin = ax.get_xlim()[0]
            xmax = ax.get_xlim()[1]
            ylabel_xpos = np.abs(xmin) / (np.abs(xmax) + np.abs(xmin))
            ylabel_halignment = 'left'

        ax.spines['left'].set_position(left_spine_pos)
        ax.yaxis.set_label_coords(ylabel_xpos, 1)
        ax.yaxis.label.set_horizontalalignment(ylabel_halignment)

    def on_ylims_change(ax):
        # update y spine position to be always at zero, top or bottom according to set limits
        # update x-label y-position
        if (np.array(ax.get_ylim()) < 0).all():  # all limits negative
            bottom_spine_pos = ('axes', 1)  # spine at the top
            xlabel_ypos = 1
        elif (np.array(ax.get_ylim()) > 0).all():  # all limits positive
            bottom_spine_pos = ('axes', 0)  # spine at the bottom
            xlabel_ypos = 0
        else:  # zero is in plot
            bottom_spine_pos = 'zero'  # spine at zero ([0, 0])
            ymin = ax.get_ylim()[0]
            ymax = ax.get_ylim()[1]
            xlabel_ypos = np.abs(ymin) / (np.abs(ymax) + np.abs(ymin))

        ax.spines['bottom'].set_position(bottom_spine_pos)
        ax.xaxis.set_label_coords(1, xlabel_ypos)

    ax.callbacks.connect('xlim_changed', on_xlims_change)
    ax.callbacks.connect('ylim_changed', on_ylims_change)

    on_xlims_change(ax)
    ax.xaxis.label.set_verticalalignment('bottom')
    ax.xaxis.label.set_horizontalalignment('right')

    ax.yaxis.label.set_rotation(0)
    on_ylims_change(ax)
    ax.yaxis.label.set_verticalalignment('top')
    ax.yaxis.label.set_horizontalalignment('left')

    ax.xaxis.label.set_fontsize(12)
    ax.yaxis.label.set_fontsize(12)


def twinx(ax, visible_spine='left'):
    """
    Create a twin Axes sharing the x-axis.

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        Existing Axes
    visible_spine: {'left', 'right'}, str, optional
        Position of the only visible axis spine

    Returns
    -------
    ax_twin: matplotlib.axes.Axes
        The newly created Axes instance

    See also
    --------
    matplotlib.axes.Axes.twinx
    twiny: Create a twin Axes sharing the y-axis.

    """
    if visible_spine in ['left', 'right']:
        # remove visible spine from hidden spine list
        hidden_spines = ['top', 'bottom', 'left', 'right']
        hidden_spines.remove(visible_spine)

        # create twiny and hide spines
        ax_twin = ax.twiny()
        for pos in hidden_spines:
            ax_twin.spines[pos].set_color('none')

        # set label position according to spine position (left/right, top)
        ax_twin.yaxis.set_label_coords(visible_spine == 'right', 1)

        return ax_twin
    else:
        # invalid keyword
        raise ValueError('Twin x-axis location must be either "left" or "right"')


def twiny(ax, visible_spine='top'):
    """
    Create a twin Axes sharing the y-axis.

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        Existing Axes
    visible_spine: {'top', 'bottom'}, str, optional
        Position of the only visible axis spine

    Returns
    -------
    ax_twin: matplotlib.axes.Axes
        The newly created Axes instance

    See also
    --------
    matplotlib.axes.Axes.twiny
    twinx: Create a twin Axes sharing the x-axis.
    """
    if visible_spine in ['top', 'bottom']:
        # remove visible spine from hidden spine list
        hidden_spines = ['top', 'bottom', 'left', 'right']
        hidden_spines.remove(visible_spine)

        # create twiny and hide spines
        ax_twin = ax.twiny()
        for pos in hidden_spines:
            ax_twin.spines[pos].set_color('none')

        # set label position according to spine position (right, bottom/top)
        ax_twin.xaxis.set_label_coords(1, visible_spine == 'top')

        return ax_twin
    else:
        # invalid keyword
        raise ValueError('Twin y-axis location must be either "top" or "bottom"')


def annotate_xtick(ax, txt, x, y=0, col='black', fs=12):
    """
    Annotate certain tick on the x-axis
    
    Parameters
    ----------
    ax: matplotlib.axes.Axes
        Current axes
    txt: string
        Annotated text
    x: float [scalar]
        x position of tick and txt
    y: float [scalar]
        y position of txt
        
    See Also
    --------
    annotate_ytick: Annotates a tick on the y-axis
    """
    txt_ret = ax.text(x, y, txt, color=col, fontsize=fs, verticalalignment='top', horizontalalignment='center',
                      bbox=dict(facecolor='white', edgecolor='none', alpha=0.75))
    line_ret, = ax.plot([x, x], [0, y], '--', color=col, lw=0.5)
    return txt_ret, line_ret


def annotate_ytick(ax, txt, x, y, col='black', fs=12):
    """
    Annotate certain tick on the y-axis
    
    Parameters
    ----------
    ax: matplotlib.axes.Axes
        Current axes
    txt: string
        Annotated text
    x: float [scalar]
        x position of tick and txt
    y: float [scalar]
        y position of txt
        
    See Also
    --------
    annotate_xtick: Annotates a tick on the x-axis
    """
    txt_ret = ax.text(x, y, txt, color=col, fontsize=fs, verticalalignment='top', horizontalalignment='center',
                      bbox=dict(facecolor='white', edgecolor='none', alpha=0.75))
    line_ret, = ax.plot([0, x], [y, y], '--', color=col, lw=0.5)
    
    return txt_ret, line_ret


def annotate_distance(ax, txt, start, stop, color='rwth:black', **kwargs):
    """
    Annotate distance between two points in given axis

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        current axis
    txt: str
        distance annotation text
    start: (float, float)
        x and y position of starting point
    stop: (float, float)
        x and y position of ending point
    color: str, optional
        annotations color
    **kwargs: dict
        all additional keyword arguments are passed to distances and texts matplotlib.axes.Axes.annotate call

    Returns
    -------
    distance: matplotlib.text.Annotation
        annotation marking distance between the two points
    text: matplotlib.text.Annotation
        annotated text

    """
    distance = ax.annotate('', xy=start, xycoords='data', xytext=stop, textcoords='data',
                           arrowprops={'arrowstyle': '|-|,widthA=0.25,widthB=0.25', 'color': color}, **kwargs)
    text = ax.annotate(txt, xy=((start[0] + stop[0]) / 2, (start[1] + stop[1]) / 2), xycoords='data',
                       xytext=(0, -2), textcoords='offset points', horizontalalignment='center',
                       verticalalignment='top',
                       bbox=wbbox, color=color, **kwargs)

    return distance, text


# Grid
def grid(ax):
    """
    Grid, but below axis
    
    Parameters
    ----------
    ax: matplotlib.axes.Axes
        Current axes
    """
    ax.grid()
    ax.set_axisbelow(True)


def update_xlim(ax, x, dx, xmax=5):
    ax.set_xlim([np.max([np.min(x), -xmax]) - dx, np.min([np.max(x), xmax]) + dx])


def update_ylim(ax, y, dy, ymax=5):
    ax.set_ylim([np.max([np.min(y), -ymax]) - dy, np.min([np.max(y), ymax]) + dy])


# Default figure sizes
fig_width = 10
fig_aspect = 16/9
landscape = {'figsize': (fig_width, fig_width/fig_aspect)}

# Styles
style_poles = {'color': 'rwth:blue', 'marker': 'x', 'mew': 2, 'ms': 5.5, 'ls': 'None'}
style_zeros = {'color': 'rwth:blue', 'marker': 'o', 'mew': 2, 'ms': 5.5, 'mfc': 'None', 'ls': 'None'}
style_graph = {'color': 'rwth:blue'}

# Widget label style (full width)
wdgtl_style = {'description_width': 'initial'}

# Axis white background
wbbox = {"facecolor": "white", "edgecolor": "None", "pad": 0}

# Propagate rwth_colors to default matplotlib colors
mcolors.get_named_colors_mapping().update(colors.rwth_colors)


# Custom stem function
def stem(ax, x, y, color='rwth:blue', **kwargs):
    container = ax.stem(x, y, use_line_collection=True, basefmt=" ", **kwargs)
    plt.setp(container, 'color', color)
    return container


def stem_set_data(container, x, y):
    tmp = [np.array([[xt, 0], [xt, yt]]) for xt, yt in zip(x, y)]
    container[1].set_segments(tmp)
    container[0].set_data(x, y)


def stem_set_xdata(container, x):
    y = container[0].get_ydata()
    stem_set_data(container, x, y)


def stem_set_ydata(container, y):
    x = container[0].get_xdata()
    stem_set_data(container, x, y)


def plot_dirac(ax, x, y, color='rwth:blue', **kwargs):
    """
    Custom dirac plot

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        Current axes
    x: array-like
        Diracs' x positions
    y: array-like
        Diracs' y-positions
    color: str, optional
        Diracs' colors
    **kwargs: dict
        all additional keyword arguments are passed to rwth_nb.plots.mpl_decorations.stem call

    Returns
    -------
    cp: instance of class matplotlib.container.StemContainer
        Container for diracs with positive weights
    cn: instance of class matplotlib.container.StemContainer
        Container for diracs with negative weights
    """
    x = np.asarray(x)
    y = np.asarray(y)

    mask = y >= 0
    xp = x[mask]
    yp = y[mask]
    if not len(xp):
        xp = np.nan*np.ones(2); yp=xp
    cp = stem(ax, xp, yp, color, markerfmt="^", **kwargs)
    
    mask = y < 0
    xn = x[mask]
    yn = y[mask]
    kwargs.pop('label', None)  # one legend label is enough

    if not len(xn):
        xn = np.nan*np.ones(2); yn=xn
    
    cn = stem(ax, xn, yn, color, markerfmt="v", **kwargs)
    
    return cp, cn


def dirac_set_data(containers, x, y):
    """
    Change data in existing dirac containers

    Parameters
    ----------
    containers: Tuple of matplotlib.container.StemContainer class
        Tuple containing both positive and negative dirac containers
    x: array-like
        New dirac x positions
    y: array-like
        New dirac weights
    """
    x = np.asarray(x)
    y = np.asarray(y)

    mask = y >= 0
    xp = x[mask]
    yp = y[mask]
    if len(xp):
        stem_set_data(containers[0], xp, yp)
    else:
        stem_set_data(containers[0], [], [])

    mask = y < 0
    xn = x[mask]
    yn = y[mask]
    if len(xn):
        stem_set_data(containers[1], xn, yn)
    else:
        stem_set_data(containers[1], [], [])


def dirac_weights(ax, x, y, weights, **kwargs):
    """
    Show diracs' weights in a plot

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        Current axes
    x: array-like or scalar
        Diracs' x positions
    y: array-like or scalar
        Diracs' weights' y-positions
    weights: array-like or scalar
        Diracs' weights
    **kwargs: dict
        all additional keyword arguments are passed to matplotlib.axes.Axes.text call
    """
    x = np.atleast_1d(x)
    y = np.atleast_1d(y)
    weights = np.atleast_1d(weights)

    for xt, yt, weight in zip(x, y, weights):
        if weight != 1:
            ax.text(xt, yt, '(' + str(weight) + ')', **kwargs)


# Laplace Region of Convergence
def plot_lroc(ax, roc, xmax=12, ymax=12):
    y1 = [-ymax, -ymax]
    y2 = [ymax, ymax]

    mask = np.isinf(roc)
    roc[mask] = np.sign(roc[mask]) * xmax
    lleft, = ax.plot([roc[0], roc[0]], [y1[0], y2[0]], ls="--", c="rwth:blue-50")
    lright, = ax.plot([roc[1], roc[1]], [y1[0], y2[0]], ls="--", c="rwth:blue-50")
    hatch = ax.fill_between(roc, y1, y2, facecolor="none", hatch="\\", edgecolor="rwth:blue-50", linewidth=0.0)

    return [lleft, lright, hatch]

def update_lroc(ax, plot, roc, xmax=12, ymax=12):
    y1 = [-ymax, -ymax]
    y2 = [ymax, ymax]

    mask = np.isinf(roc)
    roc[mask] = np.sign(roc[mask]) * xmax

    plot[0].set_xdata([roc[0], roc[0]])
    plot[1].set_xdata([roc[1], roc[1]])
    plot[2].remove()
    plot[2] = ax.fill_between(roc, y1, y2, facecolor="none", hatch="\\", edgecolor="rwth:blue-50",
                                             linewidth=0.0)

    return plot

# z Region of Convergence
def plot_zroc(ax, roc, rmax=12):
    from matplotlib.patches import Circle
    mask = np.isinf(roc)
    roc[mask] = np.sign(roc[mask]) * rmax

    # plot circles
    unitcircle = Circle((0, 0), radius=1, edgecolor="rwth:black", fill=False, linestyle='-')
    ax.add_artist(unitcircle)

    theta = np.linspace(0, 2 * np.pi, 1001)
    xs = np.outer(np.abs(roc), np.cos(theta))
    ys = np.outer(np.abs(roc), np.sin(theta))
    xs[1, :] = xs[1, ::-1]
    ys[1, :] = ys[1, ::-1]

    return ax.fill(np.ravel(xs), np.ravel(ys), facecolor="none", hatch="\\", edgecolor="rwth:blue-50", linestyle='--')[0]


def annotate_order(ax, p, ord):
    for index, order in enumerate(ord):
        if order > 1:
            ax.text(np.real(p[index]), np.imag(p[index]), '(' + str(order) + ')', color='rwth:black')
            ax.text(np.real(p[index]), -np.imag(p[index]), '(' + str(order) + ')', color='rwth:black')



import os
import requests
from IPython.lib import kernel
from notebook import notebookapp

def get_contents():
    servers = list(notebookapp.list_running_servers())
    
    token = servers[0]['token']
    api_url = servers[0]['url'] + 'api'

    connection_file_path = kernel.get_connection_file()
    connection_file = os.path.basename(connection_file_path)
    kernel_id = connection_file.split('-', 1)[1].split('.')[0]
    
    headers = {
        'Authorization': f'token {token}'
    }

    r = requests.request('GET', f'{api_url}/sessions', headers=headers)
    
    for session in r.json():
        if session['kernel']['id'] == kernel_id:
            path = session['notebook']['path']

    r = requests.request('GET', f'{api_url}/contents/{path}', headers=headers)

    return r.json()


#!/usr/bin/env python3

from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


def get_version():
    """
    Read git tag and version given by environment variable and convert it to a version number.
    The build will only be processed if you are standing at a clean tag, as this will result in 
    a clean versioning scheme.
    :return:
    """
    import subprocess

    try:
        git_describe = subprocess.check_output(["git", "describe", "--tags"])
        git_describe = git_describe.decode("utf-8").strip()

        f = open('version.txt','w')
        f.write(git_describe)
        f.close()
        # and do it one more for packaging if possible
        # f = open('src/rwth-nb/version.txt','w')
        # f.write(git_describe)
        # f.close()
    except:
        f = open('version.txt','r')
        git_describe = f.readline()
        f.close()

    version = git_describe
    split_describe = git_describe.split('-')
    if (len(split_describe) > 1):
        raise Exception("I do not want to build this package since you did not tag your current state of work.")

    return version


setup(name='rwth_nb',
      # Versions should comply with PEP440.  For a discussion on single-sourcing
      # the version across setup.py and the project code, see
      # https://packaging.python.org/en/latest/single_source_version.html
      version=get_version(),
      description='RWTH Python Library for Jupyter Notebooks',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://git.rwth-aachen.de/jupyter/rwth-nb',
      author='Christian Rohlfing, Lars Thieling, Christoph Weyer, Jens Schneider, Steffen Vogel',
      author_email='rohlfing@ient.rwth-aachen.de',
      license='MIT',
      packages=find_packages(exclude=['contrib', 'docs', 'tests']),
      install_requires=[], # todo
      zip_safe=False)



import requests
import json
import os
import datetime as dt

import rwth_nb.misc.notebook as notebook


class Submission:
    """Submit and retrieve submissions to RWTH JupyterHub Service"""

    def __init__(self, realm):
        self.realm = realm
        
        self.profile = os.environ.get('JUPYTERHUB_PROFILE')
        self.token = os.environ.get('JUPYTERHUB_API_TOKEN')

        self.service_url = 'http://proxy-public.jhub/services/submission'
        
        if self.token is None or self.profile is None:
            raise Exception('Submissions only work inside the RWTHjupyter cluster!')
    
    def url(self, user='me', ident=None):
        url = f'{self.service_url}/{self.profile}/{self.realm}'
        
        if user is not None:
            url += f'/{user}'
            
            if ident is not None:
                url += f'/{ident}'
                
        return url
    
    def headers(self):
        return {
            'Authorization': f'token {self.token}'
        }
    
    def _parse_get(self, response):
        payload = response.json()
        
        if 'data' in payload:
            for data in payload['data']:
                if 'submitted' in data:
                    data['submitted'] = dt.datetime.fromisoformat(data['submitted'])
                    
        return payload

    def submit(self, data):
        """Submit some arbitrary JSON payload"""

        r = requests.request('POST',
                             url=self.url(user=None),
                             data=json.dumps(data),
                             headers=self.headers())
        r.raise_for_status()
        
    def submit_notebook(self):
        """Submit the notebook as JSON"""

        nb = notebook.get_contents()
        
        self.submit(nb)

    def is_submitted(self):
        """Check if the user has already submitted something to this realm"""
        
        s = self.get()
        
        return len(s) > 0
    
    def get(self, user='me', ident=None, limit=None):
        """Get all submissions for a user"""

        params = {}
        if limit is not None:
            params['limit'] = limit
        
        r = requests.request('GET', url=self.url(user, ident),
                             headers=self.headers(),
                             params=params)
        r.raise_for_status()

        j = self._parse_get(r)
        
        return j['submissions']
    
    def fetch_notebook(self, user='me', ident=None, limit=None, subdir=None, prefix=None):
        subs = self.get(user, ident, limit)
        
        for sub in subs:            
            nb = sub.pop('data')
            
            realm = sub['realm']
            submitter = sub['submitter']
            id = sub['id']
            name, ext = os.path.splitext(nb['name'])
            
            # Add submission details as metadata to the notebook
            nb['content']['metadata']['submission'] = sub
            
            fn = f'{name}_{realm}_{submitter}-{id}{ext}'
            
            if prefix:
                fn = f'{prefix}_{fn}'

            if subdir:
                os.makedirs(subdir)
                fn = f'{subdir}/{fn}'
            
            with open(fn, 'w+') as nb_file:
                json.dump(nb['content'], nb_file)
                
    def fetch_all_notebooks(self):
        """Get all submissions in a realm"""

        return self.fetch_notebook(user=None)
    
    def delete(self, user='me', ident=None):
        """Delete the users submission"""
        
        r = requests.request('DELETE', url=self.url(user, ident),
                             headers=self.headers())
        r.raise_for_status()
        
    def get_all(self):
        """Get all submissions in a realm"""

        return self.get(user=None)
    
    def delete_all(self):
        """Delete all submissions"""
        
        return self.delete(user=None)